from sqlalchemy import and_
from sqlalchemy.orm import joinedload
from database import db, ITEMS_PER_PAGE
from flask_covid19.blueprints.app_all.all_model import BlueprintDateReported, BlueprintFactTable, AllImport
from flask_covid19.blueprints.app_all.all_model import BlueprintLocationGroup, BlueprintLocation, AllFlat


class WhoDateReported(BlueprintDateReported):
    __tablename__ = 'who_date_reported'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint(
            'date_reported_import_str',
            'datum',
            'year_day_of_year',
            name="uix_who_date_reported"
        ),
    )

    def __str__(self):
        return self.datum.isoformat()

    id = db.Column(db.Integer, primary_key=True)
    year_month = db.Column(db.String(255), nullable=False)
    year_week = db.Column(db.String(255), nullable=False)
    year_day_of_year = db.Column(db.String(255), nullable=False, unique=True)
    date_reported_import_str = db.Column(db.String(255), nullable=False, unique=True)
    datum = db.Column(db.Date, nullable=False, unique=True)
    year = db.Column(db.Integer, nullable=False)
    month = db.Column(db.Integer, nullable=False)
    day_of_month = db.Column(db.Integer, nullable=False)
    day_of_week = db.Column(db.Integer, nullable=False)
    week_of_year = db.Column(db.Integer, nullable=False)
    day_of_year = db.Column(db.Integer, nullable=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)


class WhoCountryRegion(BlueprintLocationGroup):
    __tablename__ = 'who_country_region'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint('location_group', name="uix_who_country_region"),
    )

    id = db.Column(db.Integer, primary_key=True)
    location_group = db.Column(db.String(255), nullable=False, unique=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)

    def __str__(self):
        return " " + self.location_group + " "


class WhoCountry(BlueprintLocation):
    __tablename__ = 'who_country'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint('location_code', 'location', name="uix_who_country"),
    )

    id = db.Column(db.Integer, primary_key=True)
    location_code = db.Column(db.String(255), unique=True, nullable=False)
    location = db.Column(db.String(255), unique=True, nullable=False)
    location_group_id = db.Column(db.Integer, db.ForeignKey('who_country_region.id'), nullable=False)
    location_group = db.relationship(
        'WhoCountryRegion',
        lazy='joined',
        cascade='save-update',
        order_by='WhoCountryRegion.location_group')
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)

    def __str__(self):
        return self.location_group.__str__() + " : "+ self.location_code + " | " + self.location

    @classmethod
    def find_by_location_code_and_location_and_location_group(cls, location_code: str, location: str, location_group: WhoCountryRegion):
        return db.session.query(cls).filter(
            and_(
                cls.location_code == location_code,
                cls.location == location,
                cls.location_group_id == location_group.id
            )
        ).one_or_none()

    @classmethod
    def get_by_location_code_and_location_and_location_group(cls, location_code: str, location: str, location_group: WhoCountryRegion):
        return db.session.query(cls).filter(
            and_(
                cls.location_code == location_code,
                cls.location == location,
                cls.location_group_id == location_group.id
            )
        ).one()

    @classmethod
    def find_by_location_code_and_location(cls, location_code: str, location: str):
        return db.session.query(cls).filter(
            and_(
                cls.location_code == location_code,
                cls.location == location
            )
        ).one_or_none()

    @classmethod
    def get_by_location_code_and_location(cls, location_code: str, location: str):
        return db.session.query(cls).filter(
            and_(
                cls.location_code == location_code,
                cls.location == location
            )
        ).one()

    @classmethod
    def find_by_location_group(cls, location_group: WhoCountryRegion):
        return db.session.query(cls).filter(
            cls.location_group == location_group
        ).order_by(cls.location).all()

    @classmethod
    def get_by_location_group(cls, location_group: WhoCountryRegion, page: int):
        return db.session.query(cls).filter(
            cls.location_group == location_group
        ).order_by(cls.location).paginate(page, per_page=ITEMS_PER_PAGE)


class WhoData(BlueprintFactTable):
    __tablename__ = 'who'
    __mapper_args__ = {'concrete': True}

    id = db.Column(db.Integer, primary_key=True)
    date_reported_id = db.Column(db.Integer, db.ForeignKey('who_date_reported.id'), nullable=False)
    date_reported = db.relationship(
        'WhoDateReported',
        lazy='joined',
        cascade='save-update',
        order_by='desc(WhoDateReported.datum)')
    location_id = db.Column(db.Integer, db.ForeignKey('who_country.id'), nullable=False)
    location = db.relationship(
        'WhoCountry',
        lazy='joined',
        cascade='save-update',
        order_by='asc(WhoCountry.location)')
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    #
    cases_new = db.Column(db.Integer, nullable=False)
    cases_cumulative = db.Column(db.Integer, nullable=False)
    deaths_new = db.Column(db.Integer, nullable=False)
    deaths_cumulative = db.Column(db.Integer, nullable=False)

    @classmethod
    def delete_by_datum(cls, datum: WhoDateReported):
        for one_who_date in cls.find_by_datum(datum):
            db.session.delete(one_who_date)
            db.session.commit()
        return None

    @classmethod
    def __query_by_date_reported(cls, date_reported: WhoDateReported):
        return db.session.query(cls).filter(
            cls.date_reported_id == date_reported.id
        ).populate_existing().options(
            joinedload(cls.location).joinedload(WhoCountry.location_group),
            joinedload(cls.date_reported)
        ).order_by(
            cls.deaths_new.desc(),
            cls.cases_new.desc(),
            cls.deaths_cumulative.desc(),
            cls.cases_cumulative.desc()
        )

    @classmethod
    def __query_by_location(cls, location: WhoCountry):
        return db.session.query(cls).filter(
            cls.location_id == location.id
        ).populate_existing().options(
            joinedload(cls.location).joinedload(WhoCountry.location_group),
            joinedload(cls.date_reported)
        )

    @classmethod
    def __query_by_date_reported_and_location(cls, date_reported: WhoDateReported, location: WhoCountry):
        return db.session.query(cls).filter(
            and_(
                cls.date_reported_id == date_reported.id,
                cls.location_id == location.id
            )
        )

    @classmethod
    def find_by_date_reported(cls, date_reported: WhoDateReported):
        return cls.__query_by_date_reported(date_reported).all()

    @classmethod
    def get_by_date_reported(cls, date_reported: WhoDateReported, page: int):
        return cls.__query_by_date_reported(date_reported).paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    def get_by_location(cls, location: WhoCountry, page: int):
        return cls.__query_by_location(location).paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    def find_by_location(cls, location: WhoCountry):
        return cls.__query_by_location(location).all()

    @classmethod
    def find_by_date_reported_and_location(cls, date_reported: WhoDateReported, location: WhoCountry):
        return cls.__query_by_date_reported_and_location(date_reported, location).one_or_none()

    @classmethod
    def get_by_date_reported_and_location(cls, date_reported: WhoDateReported, location: WhoCountry):
        return cls.__query_by_date_reported_and_location(date_reported, location).one()

    @classmethod
    def get_by_date_reported_order_by_cases_new(cls, date_reported: WhoDateReported, page):
        return db.session.query(cls).filter(
                cls.date_reported_id == date_reported.id
            ).populate_existing().options(
                joinedload(cls.location).joinedload(WhoCountry.location_group),
                joinedload(cls.date_reported)
            ).order_by(
                cls.cases_new.desc()
            ).paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    def get_by_date_reported_order_by_cases_cumulative(cls, date_reported: WhoDateReported, page):
        return db.session.query(cls).filter(
                cls.date_reported_id == date_reported.id
            ).populate_existing().options(
                joinedload(cls.location).joinedload(WhoCountry.location_group),
                joinedload(cls.date_reported)
            ).order_by(
                cls.cases_cumulative.desc()
            ).paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    def get_by_date_reported_order_by_deaths_new(cls, date_reported: WhoDateReported, page):
        return db.session.query(cls).filter(
                cls.date_reported_id == date_reported.id
            ).populate_existing().options(
                joinedload(cls.location).joinedload(WhoCountry.location_group),
                joinedload(cls.date_reported)
            ).order_by(
                cls.deaths_new.desc()
            ).paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    def get_by_date_reported_order_by_deaths_cumulative(cls, date_reported: WhoDateReported, page):
        return db.session.query(cls).filter(
                cls.date_reported_id == date_reported.id
            ).populate_existing().options(
                joinedload(cls.location).joinedload(WhoCountry.location_group),
                joinedload(cls.date_reported)
            ).order_by(
                cls.deaths_cumulative.desc()
            ).paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    def get_by_location_order_by_cases_new(cls, location: WhoCountry, page):
        return db.session.query(cls).filter(
            cls.location_id == location.id
        ).populate_existing().options(
            joinedload(cls.location).joinedload(WhoCountry.location_group),
            joinedload(cls.date_reported)
        ).order_by(
            cls.cases_new.desc()
        ).paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    def get_by_location_order_by_cases_cumulative(cls, location: WhoDateReported, page):
        return db.session.query(cls).filter(
            cls.location_id == location.id
        ).populate_existing().options(
            joinedload(cls.location).joinedload(WhoCountry.location_group),
            joinedload(cls.date_reported)
        ).order_by(
            cls.cases_cumulative.desc()
        ).paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    def get_by_location_order_by_deaths_new(cls, location: WhoDateReported, page):
        return db.session.query(cls).filter(
            cls.location_id == location.id
        ).populate_existing().options(
            joinedload(cls.location).joinedload(WhoCountry.location_group),
            joinedload(cls.date_reported)
        ).order_by(
            cls.deaths_new.desc()
        ).paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    def __query_by_location_order_by_deaths_cumulative(cls, location: WhoDateReported):
        return db.session.query(cls).filter(
            cls.location_id == location.id
        ).populate_existing().options(
            joinedload(cls.location).joinedload(WhoCountry.location_group),
            joinedload(cls.date_reported)
        ).order_by(
            cls.deaths_cumulative.desc()
        )

    @classmethod
    def get_by_location_order_by_deaths_cumulative(cls, location: WhoDateReported, page):
        return cls.__query_by_location_order_by_deaths_cumulative(location).paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    def find_by_location_order_by_deaths_cumulative(cls, location: WhoDateReported):
        return cls.__query_by_location_order_by_deaths_cumulative(location).all()

    @classmethod
    def delete_data_for_one_day(cls, date_reported: WhoDateReported):
        for data in cls.find_by_date_reported(date_reported):
            db.session.delete(data)
        db.session.delete(date_reported)
        db.session.commit()
